import { prop, getModelForClass, modelOptions, ReturnModelType, Severity } from "@typegoose/typegoose";

function generateToken(): string {
    let token = "";
    for (let i = 0; i < 16; i++) {
        token += Math.random().toString(36).substring(2, 15)
    }
    return token;
}

@modelOptions({ schemaOptions: { collection: "accounts" }, options: { customName: "Account", allowMixed: Severity.ALLOW } })
class Account {
    @prop({ required: true })
    public uid!: number;

    @prop({ required: true })
    public name!: string;

    @prop({ required: true })
    public token!: string;

    public static fromUID(this: ReturnModelType<typeof Account>, uid: string | number): Promise<Account | null> {
        return this.findOne({ uid: Number(uid) });
    }

    public static fromToken(this: ReturnModelType<typeof Account>, token: string): Promise<Account | null> {
        return this.findOne({ token });
    }

    public static fromUsername(this: ReturnModelType<typeof Account>, name: string): Promise<Account | null> {
        return this.findOne({ name });
    }

    public static async new(this: ReturnModelType<typeof Account>, name: string, uid?: string | number): Promise<Account> {
        // Get last UID from database
        const lastUID = await this.findOne().sort({ uid: -1 });
        uid ||= lastUID ? lastUID.uid + 1 : 1;

        const account = await this.fromUID(uid);
        if (account) {
            throw new Error(`Account with uid ${uid} already exists.`);
        }

        const token = generateToken();

        return this.create({ uid: Number(uid), name, token });
    }

    public static async delete(this: ReturnModelType<typeof Account>, uid: string | number): Promise<boolean> {
        const account = await this.fromUID(uid);
        if (!account) throw new Error(`Account with uid ${uid} does not exist.`);

        const rsp = await this.deleteOne({ uid: Number(uid) });

        return rsp.deletedCount === 1;
    }
}

export default getModelForClass(Account);