import { prop, getModelForClass, modelOptions, ReturnModelType, Severity, DocumentType } from "@typegoose/typegoose";
import { AvatarSkillTree, AvatarType, EquipRelic, Avatar as AvatarProto } from '../data/proto/StarRail';
import AvatarConfig from "../util/excel/AvatarConfig";
import AvatarSkillTreeConfig from "../util/excel/AvatarSkillTreeConfig";
import Logger from '../util/Logger';
import Player, { LineupI } from '../game/Player';
const c = new Logger("Avatar");

@modelOptions({ schemaOptions: { collection: 'avatars' }, options: { customName: 'Avatar', allowMixed: Severity.ALLOW } })
class Avatar {
    @prop({ required: true })
    public ownerUid!: number;

    @prop({ required: true })
    public baseAvatarId!: number;

    @prop({ default: AvatarType.AVATAR_FORMAL_TYPE })
    public avatarType!: AvatarType;

    @prop({ default: 1 })
    public level!: number;

    @prop({ default: 0 })
    public exp!: number;

    @prop({ default: 0 })
    public promotion!: number;

    @prop({ default: 0 })
    public rank!: number;

    @prop({ default: null, required: false })
    public equipmentUniqueId?: number;

    @prop({ default: [] })
    public equipRelicList!: EquipRelic[];

    @prop({ default: [] })
    public skilltreeList!: AvatarSkillTree[];

    @prop({ default: { hp: 10000, sp: 10000, satiety: 100 } })
    public fightProps!: { hp: number, sp: number, satiety: number };

    public static async getAvatars(this: ReturnModelType<typeof Avatar>, player: Player) {
        const avatars = await this.find({ ownerUid: player.uid });

        // If the player doesn't have any avatars yet, add default.
        if (avatars.length < 1) {
            let marSkills;
            let mcSkills = marSkills = new Array<AvatarSkillTree>();

            for (const skill of AvatarSkillTreeConfig.fromAvatarId(8001)) {
                mcSkills.push({
                    pointId: skill.PointID,
                    level: 1,
                });
            }

            for (const skill of AvatarSkillTreeConfig.fromAvatarId(1001)) {
                marSkills.push({
                    pointId: skill.PointID,
                    level: 1,
                });
            }

            const mc = await this.create({
                ownerUid: player.uid,
                baseAvatarId: 8001, // Avatar_PlayerBoy
                skilltreeList: mcSkills,
            });

            const mar = await this.create({
                ownerUid: player.uid,
                baseAvatarId: 1001, // Avatar_Mar_7th
                skilltreeList: marSkills,
            });

            return [mc, mar];
        } else {
            return avatars;
        }
    }

    public static async getAvatar(this: ReturnModelType<typeof Avatar>, player: Player, baseAvatarId: number, error = true) {
        const avatar = await this.findOne({ ownerUid: player.uid, baseAvatarId: baseAvatarId });

        if (!avatar) {
            if (error) c.error(`Avatar ${baseAvatarId} does not exist for player ${player.uid}. This should never happen. Check your logic at the callsite.`);
            return null;
        }
        else
            return avatar;
    }

    public static async hasAvatar(this: ReturnModelType<typeof Avatar>, player: Player, baseAvatarId: number) {
        const avatar = await this.findOne({ ownerUid: player.uid, baseAvatarId: baseAvatarId });

        return !!avatar;
    }

    public static async addAvatar(this: ReturnModelType<typeof Avatar>, player: Player, baseAvatarId: number) {
        const skillTreeData = AvatarSkillTreeConfig.fromAvatarId(baseAvatarId);

        // If the player already has this avatar, don't add it again.
        if (await this.hasAvatar(player, baseAvatarId)) throw new Error(`Player ${player.uid} already has avatar ${baseAvatarId}.`);

        const avatar = await this.create({
            ownerUid: player.uid,
            baseAvatarId: baseAvatarId,
            skilltreeList: skillTreeData.map(x => x.PointID).map(x => ({ pointId: x, level: 1 })),
        });

        return avatar;
    }

    public static async removeAvatar(this: ReturnModelType<typeof Avatar>, player: Player, baseAvatarId: number) {
        return await this.deleteMany({ ownerUid: player.uid, baseAvatarId: baseAvatarId });
    }

    public static async getAvatarsForLineup(this: ReturnModelType<typeof Avatar>, player: Player, lineup: LineupI) {
        const avatars = await this.find({ ownerUid: player.uid });

        return avatars.filter(x => lineup.avatarList?.includes(x.baseAvatarId));
    }

    public toProto(this: DocumentType<Avatar>): AvatarProto {
        return {
            baseAvatarId: this.baseAvatarId,
            exp: this.exp,
            level: this.level,
            promotion: this.promotion,
            rank: this.rank,
            equipmentUniqueId: this.equipmentUniqueId,
            equipRelicList: this.equipRelicList,
            skilltreeList: this.skilltreeList,
        }
    }

}

export default getModelForClass(Avatar);