import { prop, getModelForClass, modelOptions, ReturnModelType, DocumentType, Severity } from "@typegoose/typegoose";
import { Equipment, Item, Material, PlayerSyncScNotify, Relic } from "../data/proto/StarRail";
import Player from "../game/Player";
import ItemConfig from "../util/excel/ItemConfig";
import EquipmentConfig from "../util/excel/EquipmentConfig";

export interface PayItemData {
    id: number,
    count: number
}

interface InventoryI {
    _id: number,
    nextItemUid: number,
    materials: Material[],
    relics: Relic[],
    equipments: Equipment[]
}

@modelOptions({ schemaOptions: { collection: 'inventory' }, options: { customName: 'Inventory', allowMixed: Severity.ALLOW } })
class Inventory {
    @prop({ required: true })
    public ownerUid!: number;

    @prop({ default: 1 })
    public nextItemUid!: number;

    @prop({ default: [] })
    public materials!: Material[];

    @prop({ default: [] })
    public relics!: Relic[];

    @prop({ default: [] })
    public equipments!: Equipment[];

    public static async fromUID(this: ReturnModelType<typeof Inventory>, uid: number) {
        const inv = await this.findOne({ ownerUid: uid });
        if (inv) return inv;

        const newInv = await this.create({ ownerUid: uid });
        await newInv.save();
        return newInv;
    }

    public getItemCount(this: DocumentType<Inventory>, id: number, plr: Player): number {
        const item = this.materials.find(m => m.tid === id);
        const itemData = ItemConfig.fromId(id);
        if (!itemData) return 0;

        switch (itemData.ItemMainType) {
            case "Virtual":
                return this.getVirtualItemCount(id, plr);
            default:
                return item ? item.num ?? 0 : 0;
        }
    }

    public getVirtualItemCount(this: DocumentType<Inventory>, id: number, plr: Player) {
        // TODO: Add more virtual items.   
        switch (id) {
            case 2:
                return plr.db.basicInfo.scoin!;
        }

        return 0;
    }

    public getEquipment(this: DocumentType<Inventory>, uniqueId: number) {
        return this.equipments.find(e => e.uniqueId === uniqueId);
    }

    public async addItem(this: DocumentType<Inventory>, id: number, count: number, plr: Player): Promise<boolean> {
        const itemData = ItemConfig.fromId(id) || ItemConfig.fromEquipmentId(id) || ItemConfig.fromEidolonId(id);
        if (!itemData) return false;

        switch (itemData.ItemMainType) {
            case "Virtual":
                return await this.addVirtualItem(id, count, plr);
            case "Relic":
                return await this.addRelic(id, count, plr);
            case "Equipment":
                return await this.addEquipment(id, count, 1, 1, 1);
            case "Material":
            default:
                return await this.addMaterial(id, count, plr);
        }
    }

    public async addVirtualItem(this: DocumentType<Inventory>, id: number, count: number, plr: Player) {
        this.addMaterial(id, count, plr);

        switch (id) {
            case 1:
                plr.db.basicInfo.hcoin! += count;
                await plr.save();
                return true;
            case 2:
                plr.db.basicInfo.scoin! += count;
                await plr.save();
                return true;
            default:
                return false;
        }
    }

    public async addMaterial(this: DocumentType<Inventory>, id: number, count: number, plr: Player) {
        const item = this.materials.find(m => m.tid === id);
        const itemData = ItemConfig.fromId(id);

        if (!itemData) return false;

        if (item) {
            if (!item.num) item.num = 0;
            item.num += count;

            if (item.num < 0) {
                // Remove the item if the count is less than 0.
                this.materials.splice(this.materials.indexOf(item), 1);
            } else {
                this.materials[this.materials.indexOf(item)] = item;
            }
        } else {
            this.materials.push({ tid: id, num: count });
        }

        await this.sendMaterialUpdate(plr);
        return true;
    }

    public async addEquipment(this: DocumentType<Inventory>, id: number, count = 1, level = 1, rank = 1, promotion = 1, save = true) {
        const itemData = EquipmentConfig.fromId(id);
        if (!itemData) return false;

        for (let i = 0; i < count; i++) {
            this.equipments.push({
                tid: id,
                level,
                rank,
                promotion,
                uniqueId: this.nextItemUid++,
                isProtected: false,
            });
        }
        
        return true;
    }

    /**
     * @deprecated This method is not implemented yet.
     */
    public async addRelic(this: DocumentType<Inventory>, id: number, count: number, plr: Player, save = true) {
        return false;
    }

    public async removeItem(this: DocumentType<Inventory>, id: number, count: number, plr: Player) {
        const itemData = ItemConfig.fromId(id);
        if (!itemData) return false;

        switch (itemData.ItemMainType) {
            case "Virtual":
                return this.removeVirtualItem(id, count, plr);
            case "Material":
                return this.removeMaterial(id, count, plr);
            case "Relic":
                return this.removeRelic(id, count, plr);
            case "Equipment":
                return this.removeEquipment(id);
        }

        return false;
    }

    public async removeVirtualItem(this: DocumentType<Inventory>, id: number, count: number, plr: Player) {
        switch (id) {
            case 1:
                plr.db.basicInfo.hcoin! -= count;
                await plr.save();
                return true;
            case 2:
                plr.db.basicInfo.scoin! -= count;
                await plr.save();
                return true;
        }

        return false;
    }

    public async removeMaterial(this: DocumentType<Inventory>, id: number, count: number, plr: Player) {
        return await this.addMaterial(id, -count, plr);
    }

    public async removeEquipment(this: DocumentType<Inventory>, uniqueId: number) {
        const item = this.equipments.find(e => e.uniqueId === uniqueId);
        if (!item) return false;

        const itemData = ItemConfig.fromId(item.tid ?? 0);
        if (!itemData || itemData?.ItemMainType !== "Equipment") return false;

        this.equipments.splice(this.equipments.indexOf(item), 1);
        await this.save();
        return true;
    }

    /**
     * @deprecated This method is not implemented yet.
    */
    public async removeRelic(this: DocumentType<Inventory>, id: number, count: number, plr: Player) {
        return false;
    }

    public async sendMaterialUpdate(this: DocumentType<Inventory>, plr: Player) {
        plr.session.send(PlayerSyncScNotify, PlayerSyncScNotify.fromPartial({
            materialList: this.materials
        }));
    }

    public async sendEquipmentUpdate(this: DocumentType<Inventory>, plr: Player) {
        plr.session.send(PlayerSyncScNotify, PlayerSyncScNotify.fromPartial({
            equipmentList: this.equipments
        }));
    }

    /**
     * @deprecated This method is not implemented yet.
     */
    public async sendRelicUpdate(this: DocumentType<Inventory>, plr: Player) {
        plr.session.send(PlayerSyncScNotify, PlayerSyncScNotify.fromPartial({
            relicList: this.relics
        }));
    }

    public async payItems(this: DocumentType<Inventory>, items: PayItemData[], plr: Player) {
        // Check if the player has enough
        for (const item of items) {
            const currentCount = await this.getItemCount(item.id, plr) ?? 0;
            if (currentCount < item.count) return false;
        }

        // We have enough of everything - pay
        for (const item of items) {
            await this.removeItem(item.id, item.count, plr);
        }

        this.sendMaterialUpdate(plr);
        return true;
    }
}

export default getModelForClass(Inventory);