import { prop, getModelForClass, modelOptions, ReturnModelType, DocumentType, Severity } from "@typegoose/typegoose";
import { ExtraLineupType, HeroBasicType, PlayerBasicInfo, Vector } from "../data/proto/StarRail";
import GPlayer, { LineupI } from "../game/Player";
import Session from "../server/kcp/Session";
import Account from "./Account";
import Avatar from "./Avatar";

interface PosData {
    planeID: number;
    floorID: number;
    pos: Vector;
}

const defaultLineup: LineupI = {
    avatarList: [1001],
    extraLineupType: ExtraLineupType.LINEUP_NONE,
    index: 0,
    isVirtual: false,
    leaderSlot: 0,
    mp: 5,
    name: "Team 1",
    planeId: 10001,
}

@modelOptions({ schemaOptions: { collection: 'players' }, options: { customName: 'Player', allowMixed: Severity.ALLOW } })
class Player {
    @prop({ required: true })
    public uid!: number;

    @prop({ required: true })
    public name!: string;

    @prop({ default: false })
    public isBanned!: boolean;

    @prop({ default: HeroBasicType.BoyWarrior })
    public heroBasicType!: HeroBasicType;

    @prop({ default: { level: 70, exp: 0, stamina: 180, mcoin: 0, hcoin: 0, scoin: 0, worldLevel: 1, nickname: "Trailblazer" } })
    public basicInfo!: PlayerBasicInfo;

    @prop({ default: { curIndex: 0, lineups: { 0: defaultLineup } } })
    public lineup!: { curIndex: number, lineups: { [key: number]: LineupI } };

    @prop({ default: { planeID: 10001, floorID: 10001001, pos: { x: 0, y: 439, z: -45507 } } })
    public posData!: PosData;

    public static async fromUID(this: ReturnModelType<typeof Player>, session: Session, uid: number) {
        const db = await this.findOne({ uid });

        return db ? new GPlayer(session, db) : this.new(session, uid);
    }

    public static async fromToken(this: ReturnModelType<typeof Player>, session: Session, token: string) {
        const acc = await Account.fromToken(token);
        const db = await this.findOne({ uid: acc!.uid });

        return db ? new GPlayer(session, db) : this.new(session, acc!.uid, acc?.name);
    }

    public static async new(this: ReturnModelType<typeof Player>, session: Session, uid: number, name?: string) {
        const dataObj = {
            uid,
            name: name ?? session.account?.name ?? uid.toString(),
            lineup: { curIndex: 0, lineups: {} },
        };

        const LINEUPS = 6;
        for (let i = 0; i < LINEUPS; i++) {
            const l: LineupI = {
                avatarList: [1001],
                extraLineupType: ExtraLineupType.LINEUP_NONE,
                index: i,
                isVirtual: false,
                leaderSlot: 0,
                mp: 4,
                name: `Team ${i}`,
                planeId: 10001
            };

            l.index = i;

            // @ts-ignore
            dataObj.lineup.lineups[i] = l;
        }

        const plr = await this.create(dataObj);
        await plr.updateOne({ basicInfo: { ...plr.basicInfo, nickname: dataObj.name } });
        const player = new GPlayer(session, plr);
        await Avatar.addAvatar(player, 1001);
        await Avatar.addAvatar(player, 8001);

        return player;
    }
}

export default getModelForClass(Player);

type PlayerDoc = DocumentType<Player>;
export { PlayerDoc };