import PlayerDB, { PlayerDoc } from "../db/Player";
import Session from "../server/kcp/Session";
import { ExtraLineupType, HeroBasicType, LineupAvatar, LineupInfo, Vector } from "../data/proto/StarRail";
import Logger from "../util/Logger";
import Account from "../db/Account";
import Avatar from "../db/Avatar";
import { Scene } from "./Scene";
import Inventory from "../db/Inventory";
import { DocumentType } from "@typegoose/typegoose";
const c = new Logger("Player");

export interface LineupI {
    avatarList?: number[];
    isVirtual?: boolean;
    planeId?: number;
    mp?: number;
    leaderSlot?: number;
    index?: number;
    extraLineupType?: ExtraLineupType;
    name?: string;
}
export interface PlayerI {
    _id: number;
    name: string;
    token: string;
    banned: boolean;
    heroBasicType: HeroBasicType;
    basicInfo: {
        nickname: string;
        level: number;
        exp: number;
        stamina: number;
        mcoin: number;
        hcoin: number;
        scoin: number;
        worldLevel: number;
    }
    lineup: {
        curIndex: number;
        lineups: {
            [key: number]: LineupI;
        };
    }
    posData: {
        floorID: number;
        planeID: number;
        pos: {
            x: number,
            y: number,
            z: number
        };
    }
}

export default class Player {
    public readonly uid: number
    public readonly scene: Scene;
    private inventory!: Awaited<ReturnType<typeof Inventory.fromUID>>;

    public constructor(readonly session: Session, public readonly db: PlayerDoc) {
        this.uid = db.uid;
        this.scene = new Scene(this);
    }

    public static fromUID(session: Session, uid: number): Promise<Player | undefined> {
        return PlayerDB.fromUID(session, uid)
    }

    public static async fromToken(session: Session, token: string): Promise<Player | undefined> {
        return PlayerDB.fromToken(session, token);
    }

    public async getLineup(lineupIndex?: number): Promise<LineupInfo> {
        // Get avatar data.
        const index = lineupIndex ?? this.db.lineup.curIndex;
        const lineup = this.db.lineup.lineups[index] || this.db.lineup.lineups[0];
        const avatars = await Avatar.getAvatarsForLineup(this, lineup);

        // Construct LineupInfo.
        const lineupAvatars: LineupAvatar[] = [];
        for (let slot = 0; slot < avatars.length; slot++) {
            lineupAvatars.push({
                slot: slot,
                avatarType: avatars[slot].avatarType,
                id: avatars[slot].baseAvatarId,
                hp: avatars[slot].fightProps.hp,
                sp: avatars[slot].fightProps.sp,
                satiety: avatars[slot].fightProps.satiety
            });
        }

        return {
            ...lineup,
            index: index,
            avatarList: lineupAvatars,
            // OHIKFCKCHEF: 3, // index or sth?
            AONNGPDNBKP: 5, // technique points
            PMCLJFADCKI: 5, // technique points
        }
    }

    public setLineup(lineup: LineupInfo, index?: number, curIndex: number = this.db.lineup.curIndex) {
        this.db.lineup.lineups[index || curIndex] = {
            ...lineup,
            avatarList: lineup.avatarList!.map(x => x.id!)
        };

        this.db.lineup.curIndex = curIndex;
    }

    public async getInventory(): Promise<Awaited<ReturnType<typeof Inventory.fromUID>>> {
        // If this players inventory has not been loaded yet, do so now.
        if (!this.inventory) {
            this.inventory = await Inventory.fromUID(this.uid);
        }

        return this.inventory;
    }

    public async save() {
        const newObj = this.db.toObject();

        // @ts-ignore
        delete newObj._id;

        await this.db.updateOne(newObj);
    }
}