import _KCP from 'node-kcp-token';
import { RemoteInfo } from 'dgram';
import { resolve } from 'path';
import fs from 'fs';
import KCP from 'node-kcp-token';
import Packet, { PacketName } from './Packet';
import Logger, { VerboseLevel } from '../../util/Logger';
import defaultHandler from '../packets/PacketHandler';
import Account from '../../db/Account';
import Player from '../../game/Player';
import { PlayerKickOutScNotify, PlayerKickOutScNotify_KickType, PlayerSyncScNotify } from '../../data/proto/StarRail';
import Avatar from '../../db/Avatar';
import SRServer from './SRServer';
import { HandshakeType } from './Handshake';
import ProtoFactory, { MessageType } from '../../util/ProtoFactory';
import CmdID from '../../data/proto/CmdID';

function r(...args: string[]) {
    return fs.readFileSync(resolve(__dirname, ...args));
}
type UnWrapMessageType<T> = T extends MessageType<infer U> ? U : T;

export default class Session {
    public key: Buffer = r('./initial.key');
    public c: Logger;
    public account!: Awaited<ReturnType<typeof Account.fromUID>>;
    public player!: Player;
    public kicked = false;
    private keepAliveInterval!: NodeJS.Timeout;

    public constructor(private kcpobj: KCP.KCP, public readonly ctx: RemoteInfo, public id: string) {
        this.ctx = ctx;
        this.c = new Logger(`${this.ctx.address}:${this.ctx.port}`, 'yellow');
        this.update();
        this.keepAlive();
    }

    public inputRaw(data: Buffer) {
        if (this.kicked) return;
        this.kcpobj.input(data);
    }

    // KeepAliveNotify gets sent every 10 seconds
    public keepAlive() {
        if (this.kicked) return;
        clearTimeout(this.keepAliveInterval);
        this.keepAliveInterval = setTimeout(() => this.onTimeout(), 20000);
    }

    private onTimeout() {
        this.c.log("Disconnected due to timeout");
        this.kick(false);
    }

    public async update() {
        if (this.kicked) return;
        if (!this.kcpobj) {
            console.error("wtf kcpobj is undefined");
            console.debug(this)
            return;
        }
        const hr = process.hrtime();

        const timestamp = hr[0] * 1000000 + hr[1] / 1000;
        this.kcpobj.update(timestamp);

        let recv;
        do {
            recv = this.kcpobj.recv();
            if (!recv) break;

            this.c.verbH(recv.toString("hex"));
            if (Packet.isValid(recv)) {
                await this.handlePacket(new Packet(recv));
            } else {
                if (Logger.VERBOSE_LEVEL >= VerboseLevel.ALL) this.c.warn("Invalid packet received");
            }

        } while (recv)

        setTimeout(() => this.update(), 1);
    }

    public async handlePacket(packet: Packet) {
        if (Logger.VERBOSE_LEVEL >= VerboseLevel.WARNS) this.c.log(packet.protoName ?? packet.cmdid)
        this.c.verbL(packet.body);

        await import(`../packets/${packet.protoName}`).then(async mod => {
            await mod.default(this, packet);
        }).catch(e => {
            if (e.code === 'MODULE_NOT_FOUND') this.c.warn(`Unhandled packet: ${packet.protoName}`);
            else this.c.error(e);

            defaultHandler(this, packet);
        });
    }

    public async sync() {
        const avatars = await Avatar.getAvatars(this.player);
        const inventory = await this.player.getInventory();

        this.send(PlayerSyncScNotify, PlayerSyncScNotify.fromPartial({
            avatarSync: {
                avatarList: avatars.map(x => x.toProto())
            },
            materialList: inventory.materials,
            equipmentList: inventory.equipments,
            relicList: inventory.relics,
            basicInfo: this.player.db.basicInfo
        }));
    }

    public async send<Class extends MessageType<any>,>(type: Class, data: UnWrapMessageType<Class>) {
        const typeName = ProtoFactory.getName(type);
        const encodedBuffer = type.encode(type.fromPartial(data)).finish();
        const packet = Packet.fromEncodedBuffer(Buffer.from(encodedBuffer), typeName);

        if (Logger.VERBOSE_LEVEL >= VerboseLevel.VERBL) {
            const msg = type.decode(encodedBuffer);
            this.c.verbL(JSON.stringify(msg, null, 4));
            this.c.verbH(Buffer.from(packet).toString("hex"));
        }

        if (!encodedBuffer) this.c.error("encodedBuffer is undefined");
        if (Logger.VERBOSE_LEVEL >= VerboseLevel.WARNS) this.c.log(typeName ?? data);
        //todo: might want to regen the ts-proto types with env = node
        this.kcpobj.send(packet);
    }

    public async sendCmdID(id: CmdID) {
        const packet = Packet.fromEncodedBuffer(Buffer.allocUnsafe(0), CmdID[id] as PacketName);
        if (Logger.VERBOSE_LEVEL >= VerboseLevel.WARNS) this.c.log(CmdID[id] ?? id);
        this.kcpobj.send(packet);
    }

    public kick(hard: boolean = true, kickType = PlayerKickOutScNotify_KickType.KICK_BLACK) {
        if (hard) this.send(PlayerKickOutScNotify, {
            kickType: kickType,
            blackInfo: {
                limitLevel: 0, // Can't find anymore lol
                beginTime: BigInt(Math.round(Date.now() / 1000)),
                endTime: BigInt(Math.round(Date.now() / 1000) + 120),
                banType: 2
            }
        });

        SRServer.getInstance().handshake(HandshakeType.DISCONNECT, this.ctx);
        SRServer.getInstance().sessions.delete(this.id);
        this.kicked = true;
        clearTimeout(this.keepAliveInterval);
    }


    public sendRaw(data: Buffer) {
        if (this.kicked) return;
        this.kcpobj.send(data);
    }
}