import { AvatarExpUpCsReq, AvatarExpUpScRsp, Retcode } from "../../data/proto/StarRail";
import Avatar from "../../db/Avatar";
import { PayItemData } from "../../db/Inventory";
import AvatarConfig from "../../util/excel/AvatarConfig";
import AvatarExpItemConfig from "../../util/excel/AvatarExpItemConfig";
import AvatarPromotionConfig from "../../util/excel/AvatarPromotionConfig";
import ExpType from "../../util/excel/ExpType";
import Packet from "../kcp/Packet";
import Session from "../kcp/Session";

export default async function handle(session: Session, packet: Packet) {
    const body = packet.body as AvatarExpUpCsReq;
    const inventory = await session.player.getInventory();

    if (!body.baseAvatarId) return session.send(AvatarExpUpScRsp, { retcode: Retcode.RETCODE_RET_REQ_PARA_INVALID, returnItemList: [] });

    // Get the target avatar.
    const avatarId = body.baseAvatarId;
    const avatar = await Avatar.getAvatar(session.player, avatarId!);
    const avatarExcelData = AvatarConfig.fromId(avatarId!);
    const avatarPromotionData = AvatarPromotionConfig.fromId(avatarId, avatar.promotion);

    // Sanity checks.
    if (!avatarExcelData) return session.send(AvatarExpUpScRsp, { retcode: 1, returnItemList: [] });
    if (!avatarPromotionData) return session.send(AvatarExpUpScRsp, { retcode: 1, returnItemList: [] });

    // Determine the next level cap based on the avatar's current promotion.
    const levelCap = avatarPromotionData.MaxLevel;

    // Determine the EXP we get from the consumed items.
    let exp = 0;
    const costMaterialList = [];
    for (const item of body.itemCost!.itemList!) {
        // Determine amount of EXP given by that item.
        // We know that the cost items given in this Req will be `PileItem`s.
        const expPerItem = AvatarExpItemConfig.fromId(item.pileItem!.itemId!)?.Exp ?? 0;

        // Add EXP for the number of items consumed.
        exp += expPerItem * item.pileItem!.itemNum!;

        // Add material to cost.
        costMaterialList.push({ id: item.pileItem!.itemId, count: item.pileItem!.itemNum } as PayItemData);
    }

    // Determine cost, which is always 10% of EXP, and add to the list of cost materials.
    const coinCost = exp * 0.1;
    costMaterialList.push({ id: 2, count: coinCost } as PayItemData);

    // Try consuming materials.
    const success = await inventory.payItems(costMaterialList, session.player);
    if (!success) {
        // ToDo: Correct retcode.
        session.send(AvatarExpUpScRsp, { retcode: 1, returnItemList: [] });
        return;
    }

    await inventory.save();

    // Cost has been paid - now level up.
    let currentAvatarExp = avatar.exp + exp;
    const curExpType = ExpType.fromId(avatarExcelData.ExpGroup, avatar.level);
    if (!curExpType) return session.send(AvatarExpUpScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR, returnItemList: [] });
    let nextRequiredExp = curExpType?.Exp ?? 0;
    while (currentAvatarExp >= nextRequiredExp && avatar.level < levelCap) {
        // Increase level.
        avatar.level++;

        // Deduct EXP necessary for this level.
        currentAvatarExp -= nextRequiredExp;

        // Determine EXP necessary for the next level.
        const nextExpType = ExpType.fromId(avatarExcelData.ExpGroup, avatar.level);
        if (!nextExpType) return session.send(AvatarExpUpScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR, returnItemList: [] });
        nextRequiredExp = nextExpType?.Exp ?? 0;
    }

    // Calculate the character's new EXP and any excess EXP.
    let excessExp = 0;
    if (avatar.level == levelCap) {
        avatar.exp = 0;
        excessExp = currentAvatarExp;
    }
    else {
        avatar.exp = currentAvatarExp;
    }

    // Save.
    await avatar.save();

    // ToDo: Handle return items.

    // Done. Sync and send response.
    await session.sync();

    session.send(AvatarExpUpScRsp, {
        retcode: 0,
        returnItemList: []
    });
}