import { DoGachaCsReq, DoGachaScRsp, GachaItem, Item, ItemList, Retcode } from "../../data/proto/StarRail";
import Avatar from "../../db/Avatar";
import { PayItemData } from "../../db/Inventory";
import Gacha from "../../game/Gacha";
import BannerConfig from "../../util/Banner";
import EquipmentConfig from "../../util/excel/EquipmentConfig";
import ItemConfig from "../../util/excel/ItemConfig";
import Packet from "../kcp/Packet";
import Session from "../kcp/Session";

export default async function handle(session: Session, packet: Packet) {
    const gachaItemList: GachaItem[] = [];
    const body = packet.body as DoGachaCsReq;
    const banner = BannerConfig.fromId(body.gachaId!);
    if (!banner) return session.send(DoGachaScRsp, { retcode: Retcode.RETCODE_RET_GACHA_ID_NOT_EXIST });

    // Pay currency.
    const inventory = await session.player.getInventory();
    const success = await inventory.payItems([{ id: banner.costItemId, count: body.gachaNum } as PayItemData], session.player);

    if (!success) {
        return session.send(DoGachaScRsp, {
            retcode: 1301
        } as DoGachaScRsp);
    }

    const results = Gacha.pullMany(banner, body.gachaNum ?? 1);
    const inv = await session.player.getInventory();

    for (const id of results) {
        const itemData = EquipmentConfig.fromId(id) || ItemConfig.fromAvatarId(id) as any;
        if (!itemData) continue;

        switch (itemData.ItemType || "Equipment") {
            case "Relic":
            case "Equipment":
                const item = await inv.addEquipment(id);
                if (!item) continue;

                await inv.addItem(252, 8, session.player);
                await inv.addItem(251, 20, session.player);

                gachaItemList.push({
                    gachaItem: {
                        itemId: id,
                        level: 1,
                        num: 1,
                        rank: 1,
                    },
                    isNew: false, // TODO (?)
                    tokenItem: {
                        itemList: [{
                            itemId: 252, // Undying Starlight
                            num: 8,
                        },
                        {
                            itemId: 251, // Undying Embers
                            num: 20
                        }]
                    }
                });
                break;
            case "Material":
            case "AvatarCard":
                // Check for duplicates
                const existing = await Avatar.getAvatar(session.player, id, false);
                if (!existing) {
                    const avatar = await Avatar.addAvatar(session.player, id);
                    if (!avatar) continue;
                }

                await inv.addItem(251, 20, session.player);

                if (existing) {
                    await inv.addItem(parseInt(`1${id}`), 1, session.player);
                    await inv.addItem(252, existing.rank > 6 ? 20 : 8, session.player);
                }

                gachaItemList.push({
                    gachaItem: {
                        itemId: id,
                        num: 1,
                    },
                    isNew: !existing,
                    tokenItem: {
                        itemList: [{
                            itemId: 252, // Undying Starlight
                            num: existing ? (existing.rank > 6 ? 20 : 8) : 0,
                        },
                        {
                            itemId: 251, // Undying Embers
                            num: 20
                        }]
                    },
                    transferItemList: {
                        itemList: [{
                            itemId: parseInt(`1${id}`), // x's Eidolon
                            num: existing ? (existing.rank > 6 ? 0 : 1) : 0,
                        }]
                    }
                })
                break;
        }
    }

    // Prevent crashing the client if the gacha list is empty.
    if (gachaItemList.length === 0) return session.send(DoGachaScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR });

    session.c.debug(JSON.stringify(gachaItemList, null, 2), gachaItemList.length);
    session.send(DoGachaScRsp, {
        gachaId: body.gachaId,
        gachaNum: body.gachaNum,
        // newGachaRandom: body.gachaRandom ?? 2503,
        // newbieGachaCnt: 0,
        // todayGachaCnt: 0,
        // todayTotalGachaCnt: 0,
        gachaItemList
    } as DoGachaScRsp);
}