import { ExpUpEquipmentCsReq, ExpUpEquipmentScRsp, Retcode } from "../../data/proto/StarRail";
import { PayItemData } from "../../db/Inventory";
import EquipmentExcel from "../../util/excel/EquipmentConfig";
import EquipmentExpItemConfig from "../../util/excel/EquipmentExpItemConfig";
import EquipmentExpType from "../../util/excel/EquipmentExpType";
import EquipmentPromotionExcel from "../../util/excel/EquipmentPromotionConfig";
import Packet from "../kcp/Packet";
import Session from "../kcp/Session";

export default async function handle(session: Session, packet: Packet) {
    const body = packet.body as ExpUpEquipmentCsReq;
    const inventory = await session.player.getInventory();

    // Get the target equipment.
    const equipmentId = body.equipmentUniqueId;
    const equipment = inventory.getEquipment(equipmentId ?? 0);
    if (!equipment || !equipment.tid || !equipment.promotion) {
        session.send(ExpUpEquipmentScRsp, { retcode: Retcode.RETCODE_RET_EQUIPMENT_NOT_EXIST });
        return;
    }

    const equipmentExcelData = EquipmentExcel.fromId(equipment.tid);
    if (!equipmentExcelData) return session.send(ExpUpEquipmentScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR });

    // Determine the next level cap based on the equipment's current promotion.
    const curEquipmentPromotion = EquipmentPromotionExcel.fromId(equipment.tid, equipment.promotion!);
    if (!curEquipmentPromotion) return session.send(ExpUpEquipmentScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR });
    const levelCap = curEquipmentPromotion.MaxLevel;

    // Determine the EXP we get from the consumed items, and the coins it will cost.
    let exp = 0;
    let cost = 0;

    const costMaterialList = [];
    const costEquipmentList = []

    for (const item of body.costData!.itemList!) {
        // Determine amount of EXP given by that item.
        // If the consumed item is a PileItem, we can fetch the EXP it gives from the excels.
        if (item.pileItem) {
            const expItemData = EquipmentExpItemConfig.fromId(item.pileItem.itemId!);
            if (!expItemData) return session.send(ExpUpEquipmentScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR });

            exp += expItemData.ExpProvide * item.pileItem.itemNum!;
            cost += expItemData.CoinCost;

            costMaterialList.push({ id: item.pileItem.itemId, count: item.pileItem.itemNum } as PayItemData);
        }
        else if (item.equipmentUniqueId) {
            const consumedEquipment = inventory.getEquipment(item.equipmentUniqueId);
            const consumedEquipmentExcelData = EquipmentExcel.fromId(consumedEquipment?.tid!);
            if (!consumedEquipmentExcelData) return session.send(ExpUpEquipmentScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR });

            exp += consumedEquipmentExcelData.ExpProvide;
            cost += consumedEquipmentExcelData.CoinCost;

            costEquipmentList.push(item.equipmentUniqueId);
        }
    }

    costMaterialList.push({ id: 2, count: cost } as PayItemData);

    // Try consuming materials.
    const success = await inventory.payItems(costMaterialList, session.player);
    if (!success) {
        session.send(ExpUpEquipmentScRsp, { retcode: Retcode.RETCODE_RET_ITEM_NOT_ENOUGH, returnItemList: [] });
        return;
    }

    for (const id of costEquipmentList) {
        await inventory.removeEquipment(id);
    }

    // Cost has been paid - now level up.
    let currentEquipmentExp = equipment.exp! + exp;
    const expTypeData = EquipmentExpType.fromId(equipmentExcelData.ExpType, equipment.level ?? 1);
    if (!expTypeData) return session.send(ExpUpEquipmentScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR });
    let nextRequiredExp = expTypeData.Exp ?? 0;
    while (currentEquipmentExp >= nextRequiredExp && equipment.level! < levelCap) {
        // Increase level.
        equipment.level!++;

        // Deduct EXP necessary for this level.
        currentEquipmentExp -= nextRequiredExp;

        // Determine EXP necessary for the next level.
        const nextExpTypeData = EquipmentExpType.fromId(equipmentExcelData.ExpType, equipment.level ?? 1);
        if (!nextExpTypeData) return session.send(ExpUpEquipmentScRsp, { retcode: Retcode.RETCODE_RET_SERVER_INTERNAL_ERROR });
        nextRequiredExp = nextExpTypeData.Exp ?? 0;
    }

    // Calculate the equipment's new EXP and any excess EXP.
    let excessExp = 0;
    if (equipment.level == levelCap) {
        equipment.exp = 0;
        excessExp = currentEquipmentExp;
    }
    else {
        equipment.exp = currentEquipmentExp;
    }

    // Save.
    await inventory.save();

    // ToDo: Handle return items.

    // Done. Sync and send response.
    await session.sync();

    session.send(ExpUpEquipmentScRsp, {
        retcode: 0,
        returnItemList: []
    });
}