import Logger from "../../util/Logger";
import Account from "../../db/Account";
import Packet from "../kcp/Packet";
import Session from "../kcp/Session";
import Player from "../../game/Player";
import { PlayerGetTokenScRsp, PlayerKickOutScNotify_KickType, Retcode } from "../../data/proto/StarRail";
import SRServer from "../kcp/SRServer";
import Config from "../../util/Config";
import Whitelist from "../../util/Whitelist";
const c = new Logger("Dispatch");

interface PlayerGetTokenCsReq {
    channel_id?: number;
    account_uid?: string;
    token?: string;
    uid?: number;
    device?: string;
}

const retWarn = (msg: string) => c.warn(msg);

export default async function handle(session: Session, packet: Packet) {
    const body = packet.body as PlayerGetTokenCsReq;
    const dataObj = {
        retcode: 0,
        secretKeySeed: BigInt(0),
    } as PlayerGetTokenScRsp;

    try {
        if (!body.token) {
            dataObj.retcode = 5;
            dataObj.msg = "Token not found";
            return;
        }
        
        let account = await Account.fromToken(body.token || "");

        if (Config.AUTO_ACCOUNT && !account) {
            account = await Account.new(body.account_uid || body.token || "");
        }

        if (!account) retWarn(`Account not found with token ${body.token}`);

        const player = await Player.fromToken(session, account?.token || "");
        if (!player) retWarn(`Player not found with accountToken ${account?.token}`);
        if (!player || !account) {
            dataObj.retcode = 6;
            dataObj.msg = "Player not found";
            return;
        }

        session.account = account;
        session.player = player;
        dataObj.uid = player.db.uid;

        if (Config.GAMESERVER.MAINTENANCE && !Whitelist.Instance.isWhitelisted(player.uid ?? -1)) {
            dataObj.retcode = Retcode.RETCODE_RET_NOT_IN_WHITE_LIST;
            dataObj.msg = Config.GAMESERVER.MAINTENANCE_MSG;

            c.warn(`Suspicious login attempt from ${session.ctx.address}:${session.ctx.port}`);
            return;
        }

        // TODO: Optimise
        for (const [key, value] of SRServer.getInstance().sessions) {
            if (value.account?.uid === account.uid && key !== session.id) {
                c.warn(`Kicked duplicate session ${account.uid} (${key})`);
                value.kick(true, PlayerKickOutScNotify_KickType.KICK_SQUEEZED);
            }
        }

        if (player.db.isBanned) {
            dataObj.retcode = 1013;
            dataObj.blackInfo = {
                banType: 2,
                beginTime: BigInt(Math.floor(Date.now() / 1000)),
                endTime: BigInt(Math.floor(Date.now() / 1000) + 86400),
                limitLevel: 0,
            }
        }

        if (account.token !== body.token) {
            retWarn(`Token invalid (${session.ctx.address}:${session.ctx.port})`);
            dataObj.retcode = 1005;
            dataObj.msg = "Token invalid";
        }
    } catch (e) {
        dataObj.retcode = 2;
        c.error(e as Error);
    } finally {
        session.send(PlayerGetTokenScRsp, dataObj);
        if (dataObj.retcode !== 0) session.kick(false);
    }

}